import bs
import bsUtils
import bsMap
import bsPowerup
import bsTeamGame
import bsScoreBoard
import random
import weakref
import bsSpaz

def getGames():
    return [HockeyGame]

class PuckDeathMessage(object):
    """A puck has died."""
    def __init__(self,puck):
        self.puck = puck

class Puck(bs.Actor):

    def __init__(self,activity,retain,translate=(0,1,0)):
        bs.Actor.__init__(self,activity,retain)

        # spawn just above the provided point
        self._spawnPos = (translate[0],translate[1]+1.0,translate[2])
        self.lastPlayersToTouch = {}
        self.node = bs.newNode("prop",
                               owner=activity,
                               attrs={'model': activity._puckModel,
                                      'colorTexture': activity._puckTex,
                                      'body':'puck',
                                      'reflection':'soft',
                                      'reflectionScale':[0.2],
                                      'shadowSize': 1.0,
                                      'isAreaOfInterest':True,
                                      'translate':self._spawnPos,
                                      'materials': (bs.objectMaterial,activity._puckMaterial.materialID)
                                      },
                               delegate=self)

    def handleMessage(self,m):
        if isinstance(m,bs.DieMessage):
            bs.deleteNode(self.node)
            activity = self._activity()
            if activity: activity.handleMessage(PuckDeathMessage(self))

        # if we go out of bounds, move back to where we started...
        elif isinstance(m,bs.OutOfBoundsMessage):
            bs.setAttr(self.node,"translate",self._spawnPos)

        elif isinstance(m,bs.HitMessage):
            bs.nodeMessage(self.node,"impulse",m.pos[0],m.pos[1],m.pos[2],
                           m.velocity[0],m.velocity[1],m.velocity[2],
                           1.0*m.magnitude,1.0*m.velocityMagnitude,m.radius,0,
                           m.forceDirection[0],m.forceDirection[1],m.forceDirection[2])

            # if this hit came from a player, log them as the last to touch us
            if m.source != None:
                activity = self._activity()
                if activity:
                    if m.source in activity.players:
                        self.lastPlayersToTouch[activity.players[m.source].getTeam().teamID] = weakref.ref(activity.players[m.source])
        else:
            bs.Actor.handleMessage(self,m)


class HockeyGame(bsTeamGame.TeamGameActivity):

    @classmethod
    def getName(c):
        return 'Hockey'

    @classmethod
    def getDescription(c,sessionType):
        return 'Score some goals.'

    @classmethod
    def getSupportedSessionTypes(self):
        return ['teams']

    @classmethod
    def getSupportedMaps(self,sessionType):
        import bsMap
        return bsMap.getMapList('hockey')

    @classmethod
    def getSettings(self,sessionType):
        return [("Score to Win",{'minValue':1,'default':1,'increment':1}),
                ("Time Limit",{'choices':[('None',0),('1 Minute',60),
                                        ('2 Minutes',120),('5 Minutes',300),
                                        ('10 Minutes',600),('20 Minutes',1200)],'default':0}),
                ("Respawn Times",{'choices':[('Shorter',0.25),('Short',0.5),('Normal',1.0),('Long',2.0),('Longer',4.0)],'default':1.0})
                ]

    def __init__(self,session,level,settings):
        bsTeamGame.TeamGameActivity.__init__(self,session,level,settings)
        self._scoreBoard = bsScoreBoard.ScoreBoard(self)

        self._cheerSound = bs.getSound("cheer")
        self._chantSound = bs.getSound("crowdChant")
        self._foghornSound = bs.getSound("foghorn")
        self._swipSound = bs.getSound("swip")
        self._whistleSound = bs.getSound("refWhistle")
        self._puckModel = bs.getModel("puck")
        self._puckTex = bs.getTexture("puckColor")
        self._puckSound = bs.getSound("metalHit")

        self._puckMaterial = bs.Material('Puck Material')
        self._puckMaterial.addAction(actions=( ("partMod","friction",0.5)))
        self._puckMaterial.addAction(conditions=("dstIsMaterial",bs.pickupMaterial),
                                     actions=( ("partMod","collide",False) ) )
        self._puckMaterial.addAction(conditions=( ("srcYoungerThan",100),'and',
                                                  ("dstIsMaterial",bs.objectMaterial) ),
                                     actions=( ("nodeMod","collide",False) ) )
        self._puckMaterial.addAction(conditions=("dstIsMaterial",bs.footingMaterial),
                                     actions=(("impactSound",self._puckSound,0.2,5)))
        # keep track of which player last touched the puck
        self._puckMaterial.addAction(conditions=("dstIsMaterial",bs.playerMaterial),
                                     actions=(("call","atConnect",bs.WeakCall(self._handlePuckPlayerCollide)),))
        # we want the puck to kill powerups; not get stopped by them
        self._puckMaterial.addAction(conditions=("dstIsMaterial",bsPowerup.getFactory()._material),
                                     actions=(("partMod","physical",False),
                                              ("message","other","atConnect",bs.DieMessage())))
        self._scoreRegionMaterial = bs.Material('Hockey Score Region')
        self._scoreRegionMaterial.addAction(conditions=("dstIsMaterial",self._puckMaterial.materialID),
                                            actions=(("partMod","collide",True),
                                                     ("partMod","physical",False),
                                                     ("call","atConnect",bs.WeakCall(self._handleScore))))

    def getInstanceDescription(self):
        if self.settings['Score to Win'] == 1: return 'Score a goal.'
        else: return ('Score ${ARG1} goals.',self.settings['Score to Win'])

    def getInstanceScoreBoardDescription(self):
        if self.settings['Score to Win'] == 1: return 'score a goal'
        else: return ('score ${ARG1} goals',self.settings['Score to Win'])
        # s = self.settings['Score to Win']
        # return 'score '+ str(s)+' goal'+('s' if s > 1 else '')

    def onTransitionIn(self):
        bsTeamGame.TeamGameActivity.onTransitionIn(self, music='Hockey')

    def onBegin(self):
        bsTeamGame.TeamGameActivity.onBegin(self,nameSound=bs.getSound('announceHockey'))

        self.setupStandardTimeLimit(self.settings['Time Limit'])

        self._puckSpawnPos = self._level.getFlagPosition(None)
        self._spawnPuck()

        # set up the two score regions
        defs = self._level.defs
        self._scoreRegions = []
        self._scoreRegions.append(bs.NodeActor(self,bs.Actor.UNRETAINED,
                                                   "region",{'translate':defs.boxes["goal1"][0:3],
                                                             'scale':defs.boxes["goal1"][6:9],
                                                             'type':"box",
                                                             'materials':(self._scoreRegionMaterial.materialID,)},
                                                   name='Hockey Score Region 1'))
        self._scoreRegions.append(bs.NodeActor(self,bs.Actor.UNRETAINED,
                                                   "region",{'translate':defs.boxes["goal2"][0:3],
                                                             'scale':defs.boxes["goal2"][6:9],
                                                             'type':"box",
                                                             'materials':(self._scoreRegionMaterial.materialID,)},
                                                   name='Hockey Score Region 2'))
        self._updateScoreBoard()

        bs.playSound(self._chantSound)

    def onTeamJoined(self,team):
        team.data['score'] = 0
        self._updateScoreBoard()

    def _handlePuckPlayerCollide(self):
        try:
            puckNode,playerNode = bs.getCollideValues('sourceNode','opposingNode')
            puck = bs.getNodeDelegate(puckNode)
            player = bs.getNodeDelegate(playerNode).getPlayer()
        except Exception: player = puck = None
        if player is not None and puck is not None: puck.lastPlayersToTouch[player.getTeam().teamID] = weakref.ref(player)

    def _killPuck(self):
        self._puck = None

    def _handleScore(self):
        """ a point has been scored """

        # our puck might stick around for a second or two
        # we dont want it to be able to score again
        if self._puck.scored: return

        region = bs.getCollideValues("sourceNode")
        for i in range(len(self._scoreRegions)):
            if region == self._scoreRegions[i].node:
                break;

        scoringTeam = None
        for team in self.teams.values():
            if team.teamID == i:
                scoringTeam = team
                team.data['score'] += 1

                # tell all players to celebrate
                for player in team.players.values():
                    try: bs.nodeMessage(player.actor.node,'celebrate',2000)
                    except Exception: pass

                # if weve got the player from the scoring team that last touched us, give them points
                if scoringTeam.teamID in self._puck.lastPlayersToTouch and self._puck.lastPlayersToTouch[scoringTeam.teamID]() is not None:
                    self.scoreSet.playerScored(self._puck.lastPlayersToTouch[scoringTeam.teamID](),100,bigMessage=True)

                # end game if we won
                if team.data['score'] >= self.settings['Score to Win']:
                    self.endGame()
                    

        bs.playSound(self._foghornSound)
        bs.playSound(self._cheerSound)

        self._puck.scored = True

        # kill the puck (it'll respawn itself shortly)
        bs.gameTimer(1000,bs.WeakCall(self._killPuck))

        light = bs.newNode("light",
                           owner=self,
                           attrs={"translate": bs.getCollideValues("position"),
                                  "heightAttenuated":False,
                                  "color": (1,0,0)})
        bsUtils.animate(light,"intensity",{0:0,500:1,1000:0},loop=True)
        bs.gameTimer(1000,bs.WeakCall(bs.deleteNode,light))

        self.cameraFlash(duration=10)
        self._updateScoreBoard()

    def endGame(self):
        results = bsTeamGame.Results(scoreName='Goals')
        for t in self.teams.values(): results.setTeamScore(t,t.data['score'])
        self.end(results=results)

    def _updateScoreBoard(self):
        """ update scoreboard and check for winners """
        winScore = self.settings['Score to Win']
        for team in self.teams.values():
            self._scoreBoard.setTeamValue(team,team.data['score'],winScore)

    def handleMessage(self,m):

        # respawn dead players if they're still in the game
        if isinstance(m,bsSpaz.PlayerSpazDeathMessage):
            self.__superHandleMessage(m) # augment standard behavior
            self.respawnPlayer(m.spaz.getPlayer())
        # respawn dead pucks
        elif isinstance(m,PuckDeathMessage):
            if self.isRunning():
                self._puckRespawnTimer = bs.Timer(3000,bs.WeakCall(self._spawnPuck))
        else: self.__superHandleMessage(m)

    def __superHandleMessage(self,m):
        super(HockeyGame,self).handleMessage(m)

    def _flashPuckSpawn(self):
        light = bs.newNode("light",
                           owner=self,
                           attrs={"translate": self._puckSpawnPos,
                                  "heightAttenuated":False,
                                  "color": (1,0,0)})
        bsUtils.animate(light,"intensity",{0:0,250:1,500:0},loop=True)
        bs.gameTimer(1000,bs.WeakCall(bs.deleteNode,light))

    def _spawnPuck(self):
        bs.playSound(self._swipSound)
        bs.playSound(self._whistleSound)
        self._flashPuckSpawn()

        self._puck = Puck(self,bs.Actor.UNRETAINED,
                          translate=self._puckSpawnPos)
        self._puck.scored = False
        self._puck.lastHoldingPlayer = None
        self._puck.light = bs.newNode("light",
                                      owner=self._puck.node,
                                      attrs={"intensity":0.3,
                                             "heightAttenuated":False,
                                             "radius":0.2,
                                             "color": (0.3,0.0,1.0)})
        bs.connectAttr(self._puck.node,"translate",self._puck.light,"translate")

